// Copyright 2005, 2006 - Morten Nielsen (www.iter.dk)
//
// This file is part of SharpMap.
// SharpMap is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// SharpMap is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SharpMap; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 

using System;
using System.Collections.Generic;
using System.Text;

namespace SharpMap.Layers
{
#if !MONO
	/// <summary>
	/// ECW raster image layer
	/// </summary>
	/// <remarks>
	/// Before using the ECW raster layer, you will need to download and install the ECW SDK from
	/// <a href="http://www.ermapper.com">ER Mappers webpage</a>. Furthermore, the ECW georeference
	/// header must be correct (you can use the ECW Header Editor from ER Mapper to correct it if nessesary).
	/// <example>
	/// <code lang="C#">
	/// myMap = new SharpMap.Map(new System.Drawing.Size(500,250);
	/// SharpMap.Layers.EcwRasterLayer layEcw = new SharpMap.Layers.EcwRasterLayer("Blue Marble", @"C:\data\image.ecw");
	/// myMap.Layers.Add(layEcw);
	/// myMap.ZoomToExtents();
	/// </code>
	/// </example>
	/// </remarks>
	public class EcwRasterLayer : SharpMap.Layers.Layer, IDisposable
	{
		private SharpMap.Geometries.BoundingBox _Envelope;
		private NCSECWLib.NCSRendererClass ecw;

		/// <summary>
		/// Initializes a new ECW or JPEG2000 based Raster layer
		/// </summary>
		/// <param name="strLayerName">Name of layer</param>
		/// <param name="imageFilename">Location of image</param>
		public EcwRasterLayer(string strLayerName, string imageFilename)
		{
			this.LayerName = strLayerName;
			this.Filename = imageFilename;
			disposed = false;
			ecw = new NCSECWLib.NCSRendererClass();
			ecw.Open(this.Filename,false);
            if(ecw.CellIncrementY<0)
			    _Envelope = new SharpMap.Geometries.BoundingBox(ecw.OriginX, ecw.OriginY + ecw.Height * ecw.CellIncrementY, ecw.OriginX + ecw.Width * ecw.CellIncrementX, ecw.OriginY);
            else
                _Envelope = new SharpMap.Geometries.BoundingBox(ecw.OriginX, ecw.OriginY, ecw.OriginX + ecw.Width * ecw.CellIncrementX, ecw.OriginY + ecw.Height * ecw.CellIncrementY);
		}

		private string _Filename;
		/// <summary>
		/// Gets or sets the filename of the raster file
		/// </summary>
		public string Filename
		{
			get { return _Filename; }
			set { _Filename = value; }
		}

		#region ILayer Members

		/// <summary>
		/// Renders the layer
		/// </summary>
		/// <param name="g">Graphics object reference</param>
		/// <param name="map">Map which is rendered</param>
		public override void Render(System.Drawing.Graphics g, Map map)
		{
			if (disposed)
				throw (new ApplicationException("Error: An attempt was made to render a disposed layer"));
			if (this.Envelope.Intersects(map.Envelope))
			{
				double left = Math.Max(map.Envelope.Left, _Envelope.Left);
				double top = Math.Min(map.Envelope.Top, _Envelope.Top);
				double right = Math.Min(map.Envelope.Right, _Envelope.Right);
				double bottom = Math.Max(map.Envelope.Bottom, _Envelope.Bottom);
				System.Drawing.PointF tiepoint = map.WorldToImage(new SharpMap.Geometries.Point(left, top));
				int width = (int)Math.Round((right - left) / map.PixelSize);
				int height = (int)Math.Round((top - bottom) / map.PixelSize);
				if (width == 0 || height == 0)
					return;
				object[] ImageBands = new object[ecw.NumberOfBands];
				for (int i = 0; i < ImageBands.Length; i++)
					ImageBands[i] = i;
				object oImageBands = (object)ImageBands;

				IntPtr hdc = g.GetHdc();
				try
				{
					ecw.SetView(ecw.NumberOfBands, ref oImageBands, width, height, left, top, right, bottom);
					ecw.ReadImage(left, top, right, bottom, width, height);
					ecw.DrawImage(hdc.ToInt32(), (int)tiepoint.X, (int)tiepoint.Y, width + (int)tiepoint.X, height + (int)tiepoint.Y, left, top, right, bottom);
				}
				catch (System.Runtime.InteropServices.COMException ex)
				{
					throw (new ApplicationException("ECW Decompressor threw an exception: " + ex.Message, ex));
				}
				catch (System.Exception ex)
				{
					throw (new ApplicationException("An unknown error occured in the ECW Decompressor: " + ex.Message, ex));
				}
				finally
				{
					g.ReleaseHdc(hdc);
				}
			}
			base.Render(g, map);
		}

		/// <summary>
		/// Returns the extent of the layer
		/// </summary>
		/// <returns>Bounding box corresponding to the extent of the features in the layer</returns>
		
		public override SharpMap.Geometries.BoundingBox Envelope
		{
			get {
				return _Envelope;
			}
		}

		#endregion

		#region ICloneable Members

		/// <summary>
		/// Clones the object
		/// </summary>
		/// <returns></returns>
		public override object Clone()
		{
			throw new NotImplementedException();
		}

		#endregion


		#region Disposers and finalizers

		private bool disposed = false;

		private void Dispose(bool disposing)
		{
			if (!disposed)
			{
				if (disposing)
					if (ecw != null)
					{
						try
						{
							ecw.Close(true);
							System.Runtime.InteropServices.Marshal.ReleaseComObject(ecw);
						}
						catch { }
						finally { ecw = null; }
					}
				disposed = true;
			}
		}
		/// <summary>
		/// Disposes the EcwRasterLayer and release the raster file
		/// </summary>
		public void Dispose ()
		{
			this.Dispose (true);
			GC.SuppressFinalize(this);
		}
		/// <summary>
		/// Finalizer
		/// </summary>
		~EcwRasterLayer()
		{
			this.Dispose (true);
		}


		#endregion
	}
#endif
}
